using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Threading;
using MVR;
using MeshVR;
using System.Linq;


namespace Stopper {
	public class LightOptions : MVRScript {

        bool isSessionPlugin = false;

        AdjustLightV2 adjustLight;

        List<Light> lights = new List<Light>();

        JSONStorableFloat customShadowResolution;
        JSONStorableFloat shadowBias;
        JSONStorableFloat shadowNormalBias;
        JSONStorableFloat shadowNearPlane;

        bool wasInit = false;
        bool needUpdate = false;
        float defaultShadowBias = 0.001f;
        float defaultShadowNormalBias = 0.0f;
        float defaultShadowNearPlane = 0.1f;
		public override void Init() {
            isSessionPlugin = containingAtom.name == "CoreControl";         

            if(isSessionPlugin){
                SuperController.singleton.onAtomAddedHandlers += OnAtomAdded;
                SuperController.singleton.onAtomRemovedHandlers += OnAtomRemoved;
                foreach(Atom atom in SuperController.singleton.GetAtoms()){
                    OnAtomAdded(atom);
                }   
            } else {
                OnAtomAdded(containingAtom);
            }
           
            CreateTextField(new JSONStorableString("shadowResolution", "Note: Higher values can cause severe FPS drops. 0 defaults to standard menu setting. VeryHigh shadow quality appears to correspond to 12."), false).height = 20;
            var customResolutionSlider = AddSlider("Custom Shadow Resolution", 0f,0f, 16, (float val) => {
                int resolution = val == 0f? 0 : 1 << (int)val - 1;
                foreach(Light light in lights ){
                    light.shadowCustomResolution = resolution;
                }
            }, true);
            customShadowResolution = sliderToJSONStorableFloat[customResolutionSlider];
            
            
            CreateTextField(new JSONStorableString("shadowBias", "Shadow caster surfaces are pushed by this world-space amount away from the light, to help prevent self-shadowing artifacts. Lowering this can improve shadow resolution, but going too low can create artifacts."), false).height = 20;
            var shadowBiasSlider = AddSlider("Shadow Bias", defaultShadowBias, 0f, 1f, (float val) => {
                foreach(Light light in lights ){
                    light.shadowBias = val;
                }
            });
            shadowBiasSlider.valueFormat = "F5";
            shadowBias = sliderToJSONStorableFloat[shadowBiasSlider];

            CreateTextField(new JSONStorableString("shadowNormalBias", "Directional Light Only. Shadow caster surfaces are pushed inwards along their normals by this amount, to help prevent self-shadowing artifacts. Units of normal-based bias are expressed in terms of shadowmap texel size; typically values between 0.3-0.7 work well."), false).height = 10;
            var shadowNormalBiasSlider = AddSlider("Shadow Normal Bias", defaultShadowNormalBias, 0f, 2f, (float val) => {
                foreach(Light light in lights ){
                    light.shadowNormalBias = val;
                }
            });
            shadowNormalBiasSlider.valueFormat = "F5";
            shadowNormalBias = sliderToJSONStorableFloat[shadowNormalBiasSlider];

            CreateTextField(new JSONStorableString("shadowNearPlane", "Objects closer than this value will not cast shadows"), false).height = 10;
            var shdowNearPlaneSlider = AddSlider("Shadow Near Plane", defaultShadowNearPlane, 0f, 5f, (float val) => {
                foreach(Light light in lights ){
                    light.shadowNearPlane = val;
                }
            });
            shdowNearPlaneSlider.valueFormat = "F2";
            shadowNearPlane = sliderToJSONStorableFloat[shdowNearPlaneSlider];


            if(!isSessionPlugin){
                CreateTextField(new JSONStorableString("layerMasks", "Layers to enable for the light's culling mask. By default, everything renders on layers 0. Use the LightLayerEditor script on scene atoms to change this."), true).height = 30;
                for(int i=1;i<=31;i++) //user defined layers start with layer 8 and unity supports 31 layers
                {
                    int index = i-1;
                    var layerBool = new JSONStorableBool($"{i-1} {LayerMask.LayerToName(i)} ", true);
                    var layerToggle = CreateToggle(layerBool, true);
                    RegisterBool(layerBool);
                    layerToggle.toggle.onValueChanged.AddListener((val) => {
                        foreach(Light light in lights ){
                            int layerMask = light.cullingMask;
                            if(val) {
                                layerMask |= 1<<index;
                            } else {
                                layerMask &= ~(1<<index);
                            }
                            light.cullingMask = layerMask;
                        }
                    });
                }
            }
			wasInit = true;
		}

        public UIDynamicSlider AddSlider(string name, float defaultVal, float min, float max, JSONStorableFloat.SetFloatCallback onChange = null, bool wholeNumbers = false, bool rightSide = false){
            JSONStorableFloat storableFloat = new JSONStorableFloat(name, defaultVal, min, max, true, true);
            storableFloat.storeType = JSONStorableParam.StoreType.Full;
            UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
            uiSlider.rangeAdjustEnabled = true;
            uiSlider.slider.wholeNumbers = wholeNumbers;
            RegisterFloat(storableFloat);
            
            if(onChange != null){
                storableFloat.setCallbackFunction = onChange;
            }
            
            return uiSlider;
	    }

        private void OnAtomAdded(Atom atom) {
            if(atom.type == "InvisibleLight"){
                adjustLight = atom.GetStorableByID("Light") as AdjustLightV2;
                if(adjustLight == null){
                    return;
                }
                lights.Add(adjustLight.GetComponent<Light>());
                needUpdate = true;
            }
        }

         private void OnAtomRemoved(Atom atom) {
            if(atom.type == "InvisibleLight"){
                adjustLight = atom.GetStorableByID("Light") as AdjustLightV2;
                if(adjustLight == null){
                    return;
                }
                lights.Remove(adjustLight.GetComponent<Light>());
            }
        }
		
		private void Start(){
			
		}
		
		private void Update(){
            if(needUpdate){
                needUpdate = false;
                customShadowResolution.setCallbackFunction(customShadowResolution.val);
                shadowBias.setCallbackFunction(shadowBias.val);
                shadowNormalBias.setCallbackFunction(shadowNormalBias.val);
                shadowNearPlane.setCallbackFunction(shadowNearPlane.val);
            }
		}
	
        private void OnEnable(){
            if(wasInit) {
                customShadowResolution.SetVal(customShadowResolution.val);
                shadowBias.SetVal(shadowBias.val);
                shadowNormalBias.SetVal(shadowNormalBias.val);
                shadowNearPlane.SetVal(shadowNearPlane.val);
                needUpdate = true;
            }
        }
			
		private void OnDisable(){
            foreach(Light light in lights){
                light.shadowCustomResolution = 0;
                light.shadowBias = defaultShadowBias;
                light.shadowNormalBias = defaultShadowNormalBias;
                light.shadowNearPlane = defaultShadowNearPlane;
            }
		}
		
		private void OnDestroy(){
            SuperController.singleton.onAtomAddedHandlers -= OnAtomAdded;
            SuperController.singleton.onAtomRemovedHandlers -= OnAtomRemoved;
		}
    }
}
